(*  Title:      WellFormedness.thy
    ID:         $Id: WellFormedness.thy,v 1.4 2009/03/24 17:07:59 flokam Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2007

    Note:       Proof of WF1 and WF2, well-formedness properties 
                for ASP redustions, i.e. no creation of future 
                (resp. activity) reference to non-existing future 
                (resp. activity) by the reduction rules
*)

header {* Well-formedness for ASP *}

theory WellFormedness imports Config begin

(* A configuration is well formed if all references to futures or activities 
   map to an axisting future resp activity. A fut or act can be referenced either 
   from an active object or from a request. *)

constdefs
  wf_1a :: "configuration \<Rightarrow> bool"
  "wf_1a c == \<forall>A\<in>dom c. \<forall>E B. c.\<^sub>AA = (E::general_context)\<up>ActRef B 
                          \<longrightarrow> B \<in> dom c" 

  wf_1b  :: "configuration \<Rightarrow> bool"
  "wf_1b c == \<forall>A\<in>dom c. \<forall>fk\<in>dom (c.\<^sub>RA). \<forall>E B. c.\<^sub>FA<fk> = (E::general_context)\<up>ActRef B 
                                          \<longrightarrow> B \<in> dom c"

  wf_2a :: "configuration \<Rightarrow> bool"
  "wf_2a c == \<forall>A\<in>dom c. \<forall>E f. c.\<^sub>AA = (E::general_context)\<up>FutRef f
                          \<longrightarrow> (\<exists>B\<in>dom c. f \<in> dom (c.\<^sub>RB))" 

  wf_2b :: "configuration \<Rightarrow> bool"
  "wf_2b c == \<forall>A\<in>dom c. \<forall>fk\<in>dom (c.\<^sub>RA). \<forall>E f. c.\<^sub>FA<fk> = (E::general_context)\<up>FutRef f
                          \<longrightarrow> (\<exists>B\<in>dom c. f \<in> dom (c.\<^sub>RB))"

  wf_conf :: "configuration \<Rightarrow> bool"
  "wf_conf c == wf_1a c \<and> wf_1b c \<and> wf_2a c \<and> wf_2b c"

subsection{*General Lemmas*}
(* reduction increases the set of defined activities *)
lemma actdom_and_red_inc: "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> dom c \<subseteq> dom c'"
  by (erule zeta.induct, auto)

(* reduction increases the set of defined requests *)
lemma futdom_and_red_inc: "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> \<forall>A\<in>dom c. dom (c.\<^sub>R A) \<subseteq> dom (c'.\<^sub>RA)"
  by (erule zeta.induct, auto simp: RequestQueue_def)

(* lemma update_diff: "a \<noteq> b \<Longrightarrow> (m(a \<mapsto> f)) b = m b"
   see: ListPre.insert_select2 *)

lemma OneRequest_diff: 
  fixes c :: configuration and A :: ActivityRef and a :: sterm and s :: sterm and
        fa :: FutureRef and fb :: FutureRef and m :: futmap
  assumes "c A = Some (m,a)" "fa \<noteq> fb" "fb \<in> dom m"
  shows "(c(A \<mapsto> (m(fa \<mapsto> s),a))).\<^sub>FA<fb> = (c.\<^sub>FA<fb>)"
proof -
  from `c A = Some (m,a)` `fb \<in> dom m` have "c.\<^sub>FA<fb> = the(m fb)"
    unfolding OneRequest_def by simp
  with insert_select2[OF `fa \<noteq> fb`] `fb \<in> dom m`
  have "(m(fa \<mapsto> s)) fb = Some (c.\<^sub>FA<fb>)" by force
  moreover
  have "(c(A \<mapsto> (m(fa \<mapsto> s),a))) A = Some (m(fa \<mapsto> s),a)" by force
  ultimately show ?thesis
  using OneRequestI[of _ _ "m(fa \<mapsto> s)" a fb "c.\<^sub>FA<fb>"] by simp
qed

lemma OneRequest_diff2: 
  fixes c :: configuration and A :: ActivityRef and A' :: ActivityRef and 
        t :: "futmap \<times> sterm" and f' :: FutureRef
  assumes "A' \<in> dom c" "A \<noteq> A'"
  shows "(c(A \<mapsto> t)).\<^sub>FA'<f'> = (c.\<^sub>FA'<f'>)"
proof -
  from insert_select2[OF assms(2)] assms
  show "(c(A \<mapsto> t)).\<^sub>FA'<f'> = (c.\<^sub>FA'<f'>)" by (simp add: OneRequest_def)
qed

lemma OneRequest_eq: "(c(A \<mapsto> (m(f \<mapsto> s),a))).\<^sub>FA<f> = s"
  by (rule_tac m = "m(f \<mapsto> s)" and a = a in OneRequestI, simp+)

lemma ActiveObject_eq: "(c(A \<mapsto> (m',a'))).\<^sub>AA = a'"
  by (force simp: ActiveObject_def)

lemma ActiveObject_diff: "A \<noteq> B \<Longrightarrow> (c(B \<mapsto> t)).\<^sub>AA = (c.\<^sub>AA)"
  by (force simp: ActiveObject_def)

lemma RequestQueue_diff: "A \<noteq> B \<Longrightarrow> (c(B \<mapsto> t)).\<^sub>RA = (c.\<^sub>RA)"
  by (force simp: RequestQueue_def)

(* if an activity/future is referenced by an AO, was it the case before reduction? *)
lemma ref_and_red_ao: 
  fixes A :: ActivityRef and r :: sterm and 
        c :: configuration and c' :: configuration and E :: general_context
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "A \<in> dom c'" "c'.\<^sub>AA = E\<up>r" "isref r"
  shows "(\<exists>A E'::general_context. A \<in> dom c \<and> c.\<^sub>AA = E'\<up>r)
       \<or> (\<exists>G. G \<in> dom c 
            \<and> (\<exists>f'\<in>dom (c.\<^sub>RG). (\<exists>E'::general_context. c.\<^sub>FG<f'> = E'\<up>Active (E\<up>r))
                             \<or> (\<exists>(E'::general_context) C l E''::general_context. 
                                   c.\<^sub>FG<f'> = E'\<up>Upd (ActRef C) l (E''\<up>r))))"
proof -
  {
    fix A :: ActivityRef and c :: configuration and
        Aa :: ActivityRef and ca :: configuration and E :: general_context and
        ma :: futmap
    assume "A \<in> dom c" "c.\<^sub>AA = E\<up>r"
           "c = ca(Aa \<mapsto> (ma,ca.\<^sub>AAa))" "Aa \<in> dom ca"
    hence "\<exists>A' E'::general_context. A' \<in> dom ca \<and> (ca.\<^sub>AA') = E'\<up>r"
    proof (cases "A = Aa")
      case False with `A \<in> dom c` `(c.\<^sub>AA) = E\<up>r` `c = ca(Aa \<mapsto> (ma,ca.\<^sub>AAa))`
      show ?thesis by (simp add: ActiveObject_def, blast)
    next
      case True with `A \<in> dom c` `(c.\<^sub>AA) = E\<up>r` `c = ca(Aa \<mapsto> (ma,ca.\<^sub>AAa))`
	             `Aa \<in> dom ca`
      show ?thesis by (simp add: ActiveObject_def, blast)
    qed
  }
  note lem1 = this

  {
    fix A :: ActivityRef and c :: configuration and
        Aa :: ActivityRef and ca :: configuration and ma :: futmap and
        E :: general_context and Ba :: ActivityRef and p :: "futmap \<times> sterm"
    assume "A \<in> dom c" "c.\<^sub>AA = E\<up>r"
           "c = ca(Aa \<mapsto> (ma,ca.\<^sub>AAa), Ba \<mapsto> p)" "Aa \<in> dom ca"
    hence "(\<exists>A' E'::general_context. A' \<in> dom ca \<and> (ca.\<^sub>AA') = E'\<up>r)
          \<or>(A = Ba)"
    proof (cases "A = Ba")
      case True thus ?thesis by simp
    next
      def cb \<equiv> "ca(Aa \<mapsto> (ma, ca.\<^sub>AAa))"
      case False with `A \<in> dom c` `c = ca(Aa \<mapsto> (ma,ca.\<^sub>AAa),Ba \<mapsto> p)`
      have "A \<in> dom cb" by (simp add: cb_def)
      from `c.\<^sub>AA = E\<up>r` `c = ca(Aa \<mapsto> (ma,ca.\<^sub>AAa),Ba \<mapsto> p)` `A \<noteq> Ba`
      have "(cb.\<^sub>AA) = E\<up>r" by (simp add: cb_def ActiveObject_def)
      from lem1[OF `A \<in> dom cb` this _ `Aa \<in> dom ca`] show ?thesis by (simp add: cb_def)
    qed
  }
  note lem2 = this

(* case split *)
 from assms  show ?thesis
  proof (cases rule: zeta.cases)
    case (zeta_local Aa ca fk a') 
    from lem1[OF `A \<in> dom c'` `c'.\<^sub>AA = E\<up>r` 
                 `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fk \<mapsto> a'),ca.\<^sub>AAa))`
                 `Aa \<in> dom ca`] `c = ca`
    show ?thesis by simp
  next
    case (request_self ca fn Aa fa Ea l p) 
    from lem1[OF `A \<in> dom c'` `c'.\<^sub>AA = E\<up>r` 
                 `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>FutRef fn, 
                                          fn \<mapsto> Call (ca.\<^sub>AAa) l p),ca.\<^sub>AAa))` 
                 `Aa \<in> dom ca`] `c = ca` 
    show ?thesis by simp
  next
    case (reply Aa ca fa Ea fn Ba) thm prems(2-3) prems(6-7)
    from lem1[OF `A \<in> dom c'` `c'.\<^sub>AA = E\<up>r` 
                 `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>(ca.\<^sub>FBa<fn>)),ca.\<^sub>AAa))` 
                 `Aa \<in> dom ca`] `c = ca` 
    show ?thesis by simp
  next
    case (active G ca Aa fa Ea a) 
    note def_c' = `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>ActRef G),ca.\<^sub>AAa),G \<mapsto> (empty,a))`
    from lem2[OF `A \<in> dom c'` `c'.\<^sub>AA = E\<up>r` def_c' `Aa \<in> dom ca`] `c = ca` 
    show ?thesis proof (elim disjE, simp)
      assume "A = G" with `c'.\<^sub>AA = E\<up>r` def_c' have "a = E\<up>r" 
	by (simp add: ActiveObject_def)
      with `c = ca` `G \<notin> dom ca` `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`
	   `(ca.\<^sub>FAa<fa>) = Ea\<up>Active a` `A = G`
      show ?thesis by (simp, blast)
    qed
  next
    case (update_AO G ca Aa Ba fa Ea l p) 
    note def_c' = `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>ActRef G),ca.\<^sub>AAa), 
                           G \<mapsto> (empty, Upd (ca.\<^sub>ABa) l p))`
    from lem2[OF `A \<in> dom c'` `c'.\<^sub>AA = E\<up>r` def_c' `Aa \<in> dom ca`] `c = ca` 
    show ?thesis proof (elim disjE, simp)
      assume "A = G" with `c'.\<^sub>AA = E\<up>r` def_c' have "Upd (ca.\<^sub>ABa) l p = E\<up>r"
	by (simp add: ActiveObject_def)
      with sym[OF `c = ca`] 
	   `Aa \<in> dom ca` `Ba \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`
           `(ca.\<^sub>FAa<fa>) = Ea\<up>Upd (ActRef Ba) l p` `A = G`
	   ref_context_upd[OF `isref r` sym[OF this]]
      show ?thesis by (elim disjE exE, blast+)
    qed
  next
    case (request ca fb Aa Ba fa Ea l p) 
    note def_c' = `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>FutRef fb),ca.\<^sub>AAa),
                           Ba \<mapsto> ((ca.\<^sub>RBa)(fb \<mapsto> Call (ca.\<^sub>ABa) l p),ca.\<^sub>ABa))`
    from lem2[OF `A \<in> dom c'` `c'.\<^sub>AA = E\<up>r` def_c' `Aa \<in> dom ca`] `c = ca` 
    show ?thesis proof (elim disjE, simp)
      def cb \<equiv> "ca(Ba \<mapsto> ((ca.\<^sub>RBa)(fb \<mapsto> Call (ca.\<^sub>ABa) l p),ca.\<^sub>ABa))"
      assume "A = Ba" with `A \<in> dom c'` `c = ca` have "A \<in> dom cb" by (simp add: cb_def)
      from `c'.\<^sub>AA = E\<up>r` def_c' `A = Ba` have "cb.\<^sub>AA = E\<up>r" 
	by (simp add: cb_def ActiveObject_def)
      note lem1[OF `A \<in> dom cb` this _ `Ba \<in> dom ca`]
      from `c = ca` this[of "(ca.\<^sub>RBa)(fb \<mapsto> Call (ca.\<^sub>ABa) l p)"] show ?thesis
	by (simp add: cb_def)
    qed
  qed
qed

(* references are not created by local reduction with contexts *)
lemma clocal_and_ref[rule_format]: 
  "t \<rightarrow>\<^bsub>c\<beta>\<^esub> t' 
   \<longrightarrow> (\<forall>r E::general_context. isref r \<longrightarrow> t' = E\<up>r \<longrightarrow> (\<exists>E'::general_context. t = E'\<up>r))"
proof -
  {
    fix E' :: general_context and a :: sterm and a' :: sterm and r :: sterm and
        L :: "fVariable \<Rightarrow> bool"
    assume "finite L" and
  pred_sp: "\<forall>s p. s \<notin> L \<and> p \<notin> L \<and> s \<noteq> p 
             \<longrightarrow> (\<exists>t. (a\<^bsup>[Fvar s,Fvar p]\<^esup> \<rightarrow>\<^bsub>c\<beta>\<^esub> t 
                    \<and> (\<forall>r E::general_context. isref r \<longrightarrow> t = E\<up>r 
                        \<longrightarrow> (\<exists>E'::general_context. (a\<^bsup>[Fvar s,Fvar p]\<^esup>) = E'\<up>r))) 
                      \<and> a' = \<sigma>[s,p] t)" and
           "a' = E'\<up>r" "isref r"
    hence "\<exists>Et::general_context. a = Et\<up>r"
    proof -
      {
	fix x :: fVariable and Ex :: general_context
	from `isref r` have "Fvar x = Ex\<up>r = False"
	  by (cases rule: isref.cases, (cases Ex, auto)+)
      }note noctxt = this

      from exFresh_s_p_cof[OF `finite L`]
      obtain s p where "s \<notin> L \<and> p \<notin> L \<and> s \<noteq> p" by auto
      with pred_sp obtain t
	where  "a\<^bsup>[Fvar s,Fvar p]\<^esup> \<rightarrow>\<^bsub>c\<beta>\<^esub> t" and
	 ctxt: "\<forall>r E::general_context. isref r \<longrightarrow> t = E\<up>r 
	         \<longrightarrow> (\<exists>E'::general_context. (a\<^bsup>[Fvar s,Fvar p]\<^esup>) = E'\<up>r)" and
	       "a' = \<sigma>[s,p] t" by auto
      from this(1) cbeta_and_beta beta_lc have "lc t" by simp
      with sopen_sclose_eq_t[of t 0 s p] `a' = \<sigma>[s,p] t` `a' = E'\<up>r`
           sopen_and_copen[of r 0 "Fvar s" "Fvar p" E']
           isRef_lc[OF `isref r`]
      have "t = {0 \<rightarrow>\<^sub>c [Fvar s, Fvar p]} E'\<up>r"
	by (auto simp add: closez_def)
      with ctxt `isref r` 
      obtain Ea::general_context where "({0 \<rightarrow> [Fvar s,Fvar p]} a) = Ea\<up>r"
	by (auto simp: openz_def)
      from sopen_context_Ref[OF `isref r` this] noctxt
      show ?thesis by auto
    qed
  }note lem = this
  
  show ?thesis
  proof (rule impI, induct rule: cbeta.induct)
    case (cbeta l f T a) thus ?case
    proof (intro strip context_call_inverse)
      fix r and E :: general_context
      assume "isref r" "(the(f l)\<^bsup>[Obj f T, a]\<^esup>) = E\<up>r"
      with `l \<in> dom f` sopen_context_Ref[of r 0 "Obj f T" a "the(f l)" E]
      have "(\<exists>E::general_context. Obj f T = E\<up>r) \<or> (\<exists>E::general_context. a = E\<up>r)
	   \<or>(\<exists>E::general_context. f l = Some (E\<up>r))"
	by (auto simp: openz_def)
      thus "(\<exists>E::general_context. Obj f T = E\<up>r) \<or> (\<exists>E::general_context. a = E\<up>r)"
	by (auto intro: context_obj_inverse)
    qed
  next
    case (cupd l f T t) thus ?case
    proof (intro strip context_upd_inverse)
      fix r and E :: general_context
      assume "isref r" "Obj (f(l \<mapsto> t)) T = E\<up>r"
      from ref_context_obj[OF this(1) sym[OF this(2)]]
      obtain l' and E1::general_context 
	where "(f(l \<mapsto> t)) l' = Some (E1\<up>r)" by auto
      thus "(\<exists>E::general_context. Obj f T = E\<up>r) \<or> (\<exists>E::general_context. t = E\<up>r)"
	by (cases "l' = l", auto intro: context_obj_inverse)
    qed
  next
    case (cupdR L t t' u l) thus ?case
    proof (intro strip context_upd_inverse)
      fix r and E::general_context
      assume "isref r" "Upd u l t' = E\<up>r"
      from ref_context_upd[OF this(1) sym[OF this(2)]]
	   lem[OF `finite L` prems(2) _ `isref r`]
      show "(\<exists>E::general_context. u = E\<up>r) \<or> (\<exists>E::general_context. t = E\<up>r)"
	by auto
    qed
  next
    case (cobj l f L t t' T) thus ?case
    proof (intro strip context_obj_inverse)
      fix r and E::general_context
      assume "isref r" "Obj (f(l \<mapsto> t')) T = E\<up>r"
      from ref_context_obj[OF this(1) sym[OF this(2)]]
      obtain la and E'::general_context
	where "(f(l \<mapsto> t')) la = Some (E'\<up>r)" by auto
      with lem[OF `finite L` prems(3) _ `isref r`]
      show "\<exists>la E::general_context. (f(l \<mapsto> t)) la = Some (E\<up>r)"
	by (cases "la = l", auto)
    qed
  next
    case (cont s t E) thus ?case 
    proof (intro strip)
      fix r and Ea::general_context
      assume "isref r" "E\<up>t = Ea\<up>r" thus "\<exists>E'::general_context. E\<up>s = E'\<up>r"
      proof (cases "\<exists>E'::general_context. t = E'\<up>r")
	case True then obtain Et::general_context where "t = Et\<up>r" by auto
	with `isref r` prems(2)
	obtain Es::general_context where "s = Es\<up>r" by blast
	with `E\<up>t = Ea\<up>r` show ?thesis by (rule_tac x = "E\<circ>Es" in exI, simp)
      next
	case False with context_inc_Ref[OF `isref r` `E\<up>t = Ea\<up>r`]
	show ?thesis by simp
      qed
    qed
  qed
qed

(* idem with the classical local red *)
lemma local_and_ref: 
  "\<lbrakk> a \<rightarrow>\<^sub>\<beta> (E::general_context)\<up>r; isref r \<rbrakk> 
  \<Longrightarrow> \<exists>E'::general_context. a = E'\<up>r"
  by (auto simp: clocal_and_ref cbeta_and_beta)

(* we define the set of activities created by a reduction, ie between 2 configurations *)
consts created_act:: "configuration \<Rightarrow> configuration \<Rightarrow> ActivityRef \<Rightarrow> bool"
defs
created_def: "created_act c c' B \<equiv> c B = None \<and> (\<exists>a. c' B = Some (empty,a))"

(* we define the set of future created, inside a given activity by a reduction, 
  ie between 2 configurations *)
consts created_fut:: "configuration \<Rightarrow> configuration \<Rightarrow> ActivityRef \<Rightarrow> FutureRef \<Rightarrow> bool"
defs
fcreated_def:  "created_fut c c' A fk \<equiv> A \<in> dom c' \<and> (\<forall>B\<in>dom c. fk \<notin> dom (c.\<^sub>RB)) 
                                                   \<and> fk \<in> dom (c'.\<^sub>RA)"

(* if an activity/future is referenced by a request, was it the case before reduction? *)
lemma ref_and_red_queue:
  fixes A :: ActivityRef and r :: sterm and E :: general_context and
        c :: configuration and c' :: configuration
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "A \<in> dom c'" "fk \<in> dom(c'.\<^sub>RA)" 
          "c'.\<^sub>FA<fk> = E\<up>r" "isref r" 
  shows "(\<exists>G\<in>dom c. \<exists>E'::general_context. \<exists>fn\<in>dom (c.\<^sub>RG). c.\<^sub>FG<fn> = E'\<up>r)
        \<or>(\<exists>G\<in>dom c. \<exists>E'::general_context. c.\<^sub>AG = E'\<up>r)
        \<or>(\<exists>A. r = ActRef A \<and> created_act c c' A) 
        \<or>(\<exists>f. r = FutRef f \<and> (\<exists>B. (\<exists>l p E'::general_context. c.\<^sub>FA<fk> = E'\<up>Call (ActRef B) l p)
                                 \<and> created_fut c c' B f))"
proof -
  {
    fix A :: ActivityRef and c :: configuration and fk :: FutureRef and
        Aa :: ActivityRef and ca :: configuration and fa :: FutureRef and t :: sterm and
        E :: general_context
    assume "A \<in> dom c" "fk \<in> dom (c.\<^sub>RA)" "c.\<^sub>FA<fk> = E\<up>r"  and
   def_c': "c = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> t),ca.\<^sub>AAa))" and
           "Aa \<in> dom ca" "fa \<in> dom (ca.\<^sub>RAa)"
    hence "(A \<in> dom ca \<and> fk \<in> dom (ca.\<^sub>RA) \<and> (\<exists>E'::general_context. ca.\<^sub>FA<fk> = E'\<up>r))
          \<or>(t = E\<up>r \<and> A = Aa \<and> fk = fa)"
    proof (cases "A = Aa")
      case False with `A \<in> dom c` `fk \<in> dom (c.\<^sub>RA)` `(c.\<^sub>FA<fk>) = E\<up>r` def_c'
      show ?thesis by (simp add: RequestQueue_def OneRequest_def, blast)
    next
      case True with `(c.\<^sub>FA<fk>) = E\<up>r` def_c'
      show ?thesis
      proof (cases "fk = fa", simp add: OneRequest_def[of _ Aa fa])
	case False with `fk \<in> dom (c.\<^sub>RA)` `(c.\<^sub>FA<fk>) = E\<up>r` def_c'
	                `Aa \<in> dom ca` `A = Aa`
	show ?thesis by (simp add: RequestQueue_def OneRequest_def, blast)
      qed
    qed
  }
  note lem1 = this

  {
    fix A :: ActivityRef and c :: configuration and fk :: FutureRef and
        Aa :: ActivityRef and ca :: configuration and E :: general_context and
        fa :: FutureRef and ta :: sterm and Ba :: ActivityRef and b :: "futmap \<times> sterm"
    assume "A \<in> dom c" "fk \<in> dom (c.\<^sub>RA)" "c.\<^sub>FA<fk> = E\<up>r"  and
   def_c': "c = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> ta),ca.\<^sub>AAa), Ba \<mapsto> b)" and
           "Aa \<in> dom ca" "fa \<in> dom (ca.\<^sub>RAa)"
    hence "(A \<in> dom ca \<and> fk \<in> dom (ca.\<^sub>RA) \<and> (\<exists>E'::general_context. ca.\<^sub>FA<fk> = E'\<up>r))
          \<or>(ta = E\<up>r \<and> A = Aa \<and> fk = fa)
          \<or>(A = Ba)"
    proof (cases "A = Ba")
      case True thus ?thesis by simp
    next
      def cb \<equiv> "ca(Aa \<mapsto> ((ca .\<^sub>R Aa)(fa \<mapsto> ta), ca .\<^sub>A Aa))"
      case False with `A \<in> dom c` def_c' have "A \<in> dom cb" by (simp add: cb_def)
      from `fk \<in> dom (c.\<^sub>RA)` def_c' `A \<noteq> Ba` have "fk \<in> dom (cb.\<^sub>RA)" 
	by (simp add: cb_def RequestQueue_def)
      from `(c.\<^sub>FA<fk>) = E\<up>r` def_c' `A \<noteq> Ba` have "cb.\<^sub>FA<fk> = E\<up>r"
	by (simp add: cb_def OneRequest_def)
      note lem1[OF `A \<in> dom cb` `fk \<in> dom (cb.\<^sub>RA)` this _ `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`]
      from this[of ta] show ?thesis by (simp add: cb_def, elim disjE, blast+)
    qed
  }
  note lem2 = this

  {
    fix A :: ActivityRef and c :: configuration and fk :: FutureRef and
        Aa :: ActivityRef and ca :: configuration and fa :: FutureRef and 
        t :: sterm and t' :: sterm and Ea :: general_context and E :: general_context
    assume "Aa \<in> dom ca" "fa \<in> dom (ca.\<^sub>RAa)" "ca.\<^sub>FAa<fa> = Ea\<up>t'"
           "Ea\<up>t = E\<up>r"
    with context_inc_Ref[OF `isref r` this(4)]
    have "(Aa \<in> dom ca \<and> fa \<in> dom (ca.\<^sub>RAa) \<and> (\<exists>E1::general_context. ca.\<^sub>FAa<fa> = E1\<up>r))
         \<or>(\<exists>E2::general_context. t = E2\<up>r)"
      by (elim disjE, simp+)
  }
  note lem3 = this

(* case split *)
  from assms show ?thesis
  proof (cases rule: zeta.cases)
    case (zeta_local Aa ca fa a')
    from `ca.\<^sub>FAa<fa> \<rightarrow>\<^sub>\<beta> a'` sym[OF `c = ca`] 
         lem1[OF `A \<in> dom c'` `fk \<in> dom (c'.\<^sub>RA)` `(c'.\<^sub>FA<fk>) = E\<up>r`
                 `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> a'),ca.\<^sub>AAa))` `Aa \<in> dom ca`
                 `fa \<in> dom (ca.\<^sub>RAa)`]
    show ?thesis
    proof (elim disjE conjE, blast)
      assume "ca.\<^sub>FAa<fa> \<rightarrow>\<^sub>\<beta> a'" "a' = E\<up>r"
      with local_and_ref[OF _ `isref r`]
      obtain Ec::general_context where "ca.\<^sub>FAa<fa> = Ec\<up>r" by auto
      with `c = ca` `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)` show ?thesis by blast
    qed
  next
    case (reply Aa ca fa Ea fn Ba)
    from sym[OF `c = ca`] 
         lem1[OF `A \<in> dom c'` `fk \<in> dom (c'.\<^sub>RA)` `(c'.\<^sub>FA<fk>) = E\<up>r`
                 `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>(ca.\<^sub>FBa<fn>)),ca.\<^sub>AAa))`
                 `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`]
    show ?thesis
    proof (elim disjE conjE, blast)
      assume "Ea\<up>(ca.\<^sub>FBa<fn>) = E\<up>r"
      from `c = ca` `Ba \<in> dom ca` `fn \<in> dom (ca.\<^sub>RBa)`
	   lem3[OF `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)` 
	           `(ca.\<^sub>FAa<fa>) = Ea\<up>FutRef fn` this]
      show ?thesis by (elim disjE, blast+)
    qed
  next
    case (active G ca Aa fa Ea a)
    note def_c' = `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>ActRef G),ca.\<^sub>AAa),G \<mapsto> (empty,a))`
    from sym[OF `c = ca`] 
         lem2[OF `A \<in> dom c'` `fk \<in> dom (c'.\<^sub>RA)` `(c'.\<^sub>FA<fk>) = E\<up>r` 
                 def_c' `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`]
    show ?thesis
    proof (elim disjE conjE, blast)
      assume "Ea\<up>ActRef G = E\<up>r"
      from lem3[OF `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)` 
	           `(ca.\<^sub>FAa<fa>) = Ea\<up>Active a` this] `c = ca`
      show ?thesis
      proof (elim disjE exE, blast)
	fix E2 :: general_context assume "ActRef G = E2\<up>r"
	with `isref r` show ?thesis
	proof (cases rule: isref.cases)
	  case ref_FutRef with `ActRef G = E2\<up>r` show ?thesis 
	    by (auto simp: diff_futactrefs)
	next
	  case (ref_ActRef B) with context_actrefs[OF `ActRef G = E2\<up>r`]
	  have "G = B" by simp
	  with created_def[of c c' B] `r = ActRef B` `c = ca` def_c' `G \<notin> dom ca` 
	  show ?thesis by (simp, force)
	qed
      qed
    next
      assume "A = G" with `fk \<in> dom (c'.\<^sub>RA)` def_c' show ?thesis 
	by (simp add: RequestQueue_def) (* contradiction *)
    qed
  next
    case (update_AO G ca Aa Ba fa Ea l p)
    note def_c' = `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>ActRef G),ca.\<^sub>AAa), 
                           G \<mapsto> (empty,Upd (ca.\<^sub>ABa) l p))`
    from sym[OF `c = ca`] 
         lem2[OF `A \<in> dom c'` `fk \<in> dom (c'.\<^sub>RA)` `(c'.\<^sub>FA<fk>) = E\<up>r` 
                 def_c' `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`] show ?thesis
    proof (elim disjE conjE, blast)
      assume "Ea\<up>ActRef G = E\<up>r"
      from lem3[OF `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)` 
	           `(ca.\<^sub>FAa<fa>) = Ea\<up>Upd (ActRef Ba) l p` this] `c = ca`
      show ?thesis
      proof (elim disjE exE, blast)
	fix E2 :: general_context assume "ActRef G = E2\<up>r"
	with `isref r` show ?thesis
	proof (cases rule: isref.cases)
	  case ref_FutRef with `ActRef G = E2\<up>r` show ?thesis 
	    by (auto simp: diff_futactrefs)
	next
	  case (ref_ActRef B) with context_actrefs[OF `ActRef G = E2\<up>r`]
	  have "G = B" by simp
	  with created_def[of c c' B] `c = ca` def_c' 
	       `G \<notin> dom ca` `r = ActRef B`
	  show ?thesis by (simp, force)
	qed
      qed
    next
      assume "A = G" with `fk \<in> dom (c'.\<^sub>RA)` def_c' show ?thesis 
	by (simp add: RequestQueue_def) (* contradiction *)
    qed
  next
    case (request ca fb Aa Ba fa Ea l p)
    note def_c' = `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>FutRef fb),ca.\<^sub>AAa),
                           Ba \<mapsto> ((ca.\<^sub>RBa)(fb \<mapsto> Call (ca.\<^sub>ABa) l p), ca.\<^sub>ABa))`
    from sym[OF `c = ca`]
         lem2[OF `A \<in> dom c'` `fk \<in> dom (c'.\<^sub>RA)` `(c'.\<^sub>FA<fk>) = E\<up>r`
                 def_c' `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`] 
    show ?thesis
    proof (elim disjE conjE, blast)
      assume "Ea\<up>FutRef fb = E\<up>r" "A = Aa" "fk = fa"
      from `c = ca` lem3[OF `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`
	                    `(ca.\<^sub>FAa<fa>) = Ea\<up>Call (ActRef Ba) l p` this(1)] 
      show ?thesis
      proof (elim disjE exE, blast)
	fix E2 :: general_context assume "FutRef fb = E2\<up>r"
	from `isref r` show ?thesis
	proof (cases rule: isref.cases)
	  case ref_ActRef with `FutRef fb = E2\<up>r` show ?thesis
	    by (auto simp: diff_actfutrefs)
	next
	  case (ref_FutRef f)
	  with context_futrefs[OF `FutRef fb = E2\<up>r`] have "fb = f" by simp
	  with `c = ca` def_c' `\<forall>A\<in>dom ca. fb \<notin> dom (ca.\<^sub>RA)` fcreated_def[of c c' Ba f]
	  have "created_fut c c' Ba f" by (simp add: RequestQueue_def)
	  moreover
	  from `c = ca` `(ca.\<^sub>FAa<fa>) = Ea\<up>Call (ActRef Ba) l p`
	       `A = Aa` `fk = fa`
	  have "c.\<^sub>FA<fk> = Ea\<up>Call (ActRef Ba) l p"
	    by simp
	  ultimately show ?thesis using `r = FutRef f` by (intro disjI2, simp, blast)
	qed
      qed
    next
      def cb \<equiv> "ca(Ba \<mapsto> ((ca.\<^sub>RBa)(fb \<mapsto> Call (ca.\<^sub>ABa) l p),ca.\<^sub>ABa))"
      assume "A = Ba" with `fk \<in> dom (c'.\<^sub>RA)` def_c' have "fk \<in> dom (cb.\<^sub>RBa)"
	by (simp add: cb_def RequestQueue_def)
      from `(c'.\<^sub>FA<fk>) = E \<up> r` def_c' `A = Ba` have "cb.\<^sub>FBa<fk> = E\<up>r"
	by (simp add: cb_def OneRequest_def)
      show ?thesis proof (cases "fk = fb")
	case True with `cb.\<^sub>FBa<fk> = E\<up>r` have "Call (ca.\<^sub>ABa) l p = E\<up>r"
	  by (simp add: cb_def OneRequest_def)
	from ref_context_call[OF `isref r` sym[OF this]] 
	     sym[OF `c = ca`] `Ba \<in> dom ca` 
	show ?thesis
	proof (elim disjE exE, blast)
	  fix Ep :: general_context assume "p = Ep\<up>r"
	  with `(ca.\<^sub>FAa<fa>) = Ea\<up>Call (ActRef Ba) l p` 
	  have "ca.\<^sub>FAa<fa> = (Ea\<circ>(cCallR (ActRef Ba) l Ep))\<up>r"
	    by simp
	  with `c = ca` `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)` show ?thesis by blast
	qed
      next
	case False with `Ba \<in> dom ca` `fk \<in> dom (cb.\<^sub>RBa)` `cb.\<^sub>FBa<fk> = E\<up>r` 
	                `A = Ba` `c = ca` 
	show ?thesis
	  by (simp add: cb_def RequestQueue_def OneRequest_def, blast)
      qed
    qed
  next
    case (request_self ca fn Aa fa Ea l p) 
    note def_c' = `c' = ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>FutRef fn, 
                                           fn \<mapsto> Call (ca.\<^sub>AAa) l p),ca.\<^sub>AAa))`
    show ?thesis
    proof (cases "A = Aa")
      case True thus ?thesis proof (cases "fk = fn")
	def cb \<equiv> "ca(Aa \<mapsto> ((ca.\<^sub>RAa)(fa \<mapsto> Ea\<up>FutRef fn), ca.\<^sub>AAa))"
	from `A \<in> dom c'` def_c' have "A \<in> dom cb" by (simp add: cb_def)
	case False with `fk \<in> dom (c'.\<^sub>RA)` def_c' have "fk \<in> dom (cb.\<^sub>RA)"
	  by (auto simp: RequestQueue_def cb_def)
	from `(c'.\<^sub>FA<fk>) = E\<up>r` def_c' `A = Aa` `fk \<noteq> fn`
	have "cb.\<^sub>FA<fk> = E\<up>r" 
	  by (auto simp: OneRequest_def cb_def)
	note lem1[OF `A \<in> dom cb` `fk \<in> dom (cb.\<^sub>RA)` this _ `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`]
	from this[of "Ea\<up>FutRef fn"] sym[OF `c = ca`] show ?thesis 
	proof (simp add: cb_def, elim disjE conjE, blast)
	  assume "Ea\<up>FutRef fn = E\<up>r" "fk = fa"
	  from `c = ca` lem3[OF `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)`
	                        `(ca.\<^sub>FAa<fa>) = Ea\<up>Call (ActRef Aa) l p` this(1)] 
	  show ?thesis
	  proof (elim disjE exE, blast)
	    fix E2 :: general_context assume "FutRef fn = E2\<up>r"
	    with `isref r` show ?thesis
	    proof (cases rule: isref.cases)
	      case ref_ActRef with `FutRef fn = E2\<up>r` show ?thesis
		by (auto simp: diff_actfutrefs)
	    next
	      case (ref_FutRef f)
	      with context_futrefs[OF `FutRef fn = E2\<up>r`] have "fn = f" by simp
	      with `c = ca` def_c' `\<forall>A\<in>dom ca. fn \<notin> dom (ca.\<^sub>RA)` fcreated_def[of c c' Aa f]
	      have "created_fut c c' Aa f" by (simp add: RequestQueue_def)
	      moreover
	      from `c = ca` `(ca.\<^sub>FAa<fa>) = Ea\<up>Call (ActRef Aa) l p`
		   `A = Aa` `Ea\<up>FutRef fn = E\<up>r` `fk = fa`
	      have "c.\<^sub>FA<fk> = Ea\<up>Call (ActRef Aa) l p"	by simp
	      ultimately show ?thesis using `r = FutRef f` by (intro disjI2, simp, blast)
	    qed
	  qed
	qed
      next
	case True with `(c'.\<^sub>FA<fk>) = E\<up>r` def_c' `A = Aa`
	have "Call (ca.\<^sub>AAa) l p = E\<up>r" by (simp add: OneRequest_def)
	from ref_context_call[OF `isref r` sym[OF this]]
	     `c = ca` `Aa \<in> dom ca` 
	show ?thesis 
	proof (elim disjE exE, blast)
	  fix Ep :: general_context assume "p = Ep\<up>r"
	  with `(ca.\<^sub>FAa<fa>) = Ea\<up>Call (ActRef Aa) l p` 
	  have "ca.\<^sub>FAa<fa> = (Ea\<circ>(cCallR (ActRef Aa) l Ep))\<up>r"
	    by simp
	  with `c = ca` `Aa \<in> dom ca` `fa \<in> dom (ca.\<^sub>RAa)` show ?thesis by blast
	qed
      qed
    next 
      case False with `A \<in> dom c'` `fk \<in> dom (c'.\<^sub>RA)` `(c'.\<^sub>FA<fk>) = E\<up>r` 
	              `c = ca` def_c' 
      show ?thesis 
	by (simp add: OneRequest_def RequestQueue_def, blast)
    qed
  qed
qed

lemma fref_and_red_queue:
  fixes A :: ActivityRef and f :: FutureRef and E :: general_context and
        c :: configuration and c' :: configuration
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "A \<in> dom c'" "fk \<in> dom(c'.\<^sub>RA)" "c'.\<^sub>FA<fk> = E\<up>FutRef f" 
  shows "(\<exists>G\<in>dom c. \<exists>E'::general_context. \<exists>fn\<in>dom (c.\<^sub>RG). c.\<^sub>FG<fn> = E'\<up>FutRef f)
        \<or>(\<exists>G\<in>dom c. \<exists>E'::general_context. c.\<^sub>AG = E'\<up>FutRef f)
        \<or>(\<exists>B. (\<exists>l p E'::general_context. c.\<^sub>FA<fk> = E'\<up>Call (ActRef B) l p)
                                       \<and> created_fut c c' B f)"
proof -
  from ref_and_red_queue[OF assms] show ?thesis by simp
qed

lemma aref_and_red_queue:
  fixes A :: ActivityRef and B :: ActivityRef and E :: general_context and
        c :: configuration and c' :: configuration
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "A \<in> dom c'" "fk \<in> dom(c'.\<^sub>RA)" "c'.\<^sub>FA<fk> = E\<up>ActRef B" 
  shows "(\<exists>G\<in>dom c. \<exists>E'::general_context. \<exists>fn\<in>dom (c.\<^sub>RG). c.\<^sub>FG<fn> = E'\<up>ActRef B )
        \<or>(\<exists>G\<in>dom c. \<exists>E'::general_context. c.\<^sub>AG = E'\<up>ActRef B)
        \<or>(created_act c c' B)"
proof -
  from ref_and_red_queue[OF assms] show ?thesis by simp
qed


(* ----------------------    W F 1 ---------------------*)

subsection{* WF1 *}

(* a more precise lemma on the set of defined acts after reduction *)
lemma actdom_and_red_exact: 
  "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> dom c = dom c' 
             \<or> (\<exists>!G. G \<in> (dom c'-dom c) \<and> (\<exists> a. (c' G) = Some (empty, a)))"
proof -
  {
    fix A :: ActivityRef and c :: configuration and t :: "futmap \<times> sterm"
    assume "A \<in> dom c" hence "dom c = dom (c(A \<mapsto> t))" by auto
  }
  note domlem1 = this

  {
    fix A :: ActivityRef and B :: ActivityRef and c :: configuration and 
        t :: "futmap \<times> sterm" and a :: sterm
    assume "B \<notin> dom c" "A \<in> dom c"
    from domlem1[OF `A \<in> dom c`] have "dom c = dom (c(A \<mapsto> t))" by assumption
    with `B \<notin> dom c` have "dom (c(A \<mapsto> t, B \<mapsto> (empty,a))) - dom c = {B}"
      by force
    moreover
    have "\<exists>a'. (c(A \<mapsto> t,B \<mapsto> (empty,a))) B = Some (empty,a')"
      by (rule_tac x = a in exI, simp)
    ultimately 
    have "\<exists>!G. G \<in> dom (c(A \<mapsto> t, B \<mapsto> (empty,a))) - dom c
              \<and> (\<exists>a'. (c(A \<mapsto> t, B \<mapsto> (empty,a))) G = Some (empty,a'))"
      by (rule_tac a = B in ex1I, auto)
  }
  note domlem2 = this

(* induction *)
  show "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> dom c = dom c' 
                  \<or> (\<exists>!G. G \<in> (dom c'-dom c) \<and> (\<exists> a. (c' G) = Some (empty, a)))"
  proof (induct c c' rule: zeta.induct)
    case (zeta_local A c fk a') thus ?case
      by (simp only: domlem1[OF `A \<in> dom c`] disjI1)
  next
    case (reply A c fk E fi B) thus ?case
      by (simp only: domlem1[OF `A \<in> dom c`] disjI1)
  next
    case (request_self c fn A fk E l p) thus ?case
      by (simp only: domlem1[OF `A \<in> dom c`] disjI1)
  next
    case (update_AO G c A B fk E l p) thus ?case
      by (simp only: domlem2[OF `G \<notin> dom c` `A \<in> dom c`] disjI2)
  next
    case (active G c A fk E a) thus ?case
      by (simp only: domlem2[OF `G \<notin> dom c` `A \<in> dom c`] disjI2)
  next
    case (request c fn A B fk E l p)
    from domlem1[OF `A \<in> dom c`] `B \<in> dom c`
         domlem1[of B  "c(A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> E\<up>FutRef fn),c.\<^sub>AA))"]
    show ?case by (intro disjI1, blast)
  qed
qed

(* reduction maintans WF1a+b *)
lemma subject_reduction_WF1[rule_format]: 
  "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> (wf_1a c \<and> wf_1b c \<longrightarrow> wf_1a c' \<and> wf_1b c')"
  unfolding wf_1a_def wf_1b_def
proof (rule impI, elim conjE, intro conjI strip)
  fix c :: configuration and c' :: configuration and E :: general_context and
      A :: ActivityRef and B :: ActivityRef
  assume "c \<rightarrow>\<^sub>\<zeta> c'" and
    act: "\<forall>Aa\<in>dom c. \<forall>(Ea::general_context) Ba. (c.\<^sub>AAa) = Ea\<up>ActRef Ba \<longrightarrow> Ba \<in> dom c" and
    fut: "\<forall>Aa\<in>dom c. \<forall>fk\<in>dom (c.\<^sub>RAa). \<forall>(Ea::general_context) Ba. c.\<^sub>FAa<fk> = Ea\<up>ActRef Ba
           \<longrightarrow> Ba \<in> dom c" and "A \<in> dom c'" "c'.\<^sub>AA = E\<up>ActRef B"
  from ref_and_red_ao[OF `c \<rightarrow>\<^sub>\<zeta> c'` `A \<in> dom c'` `(c'.\<^sub>AA) = E\<up>ActRef B`] 
  show "B \<in> dom c'"
  proof (simp, elim disjE conjE exE bexE)
    fix A' :: ActivityRef and E' :: general_context
    assume "A' \<in> dom c" "c.\<^sub>AA' = E'\<up>ActRef B" with act show ?thesis
      by (simp add: rev_subsetD[OF _ actdom_and_red_inc[OF `c \<rightarrow>\<^sub>\<zeta> c'`]]) 
  next
    fix G :: ActivityRef and f' :: FutureRef and E' :: general_context
    assume "G \<in> dom c" "f' \<in> dom (c.\<^sub>RG)" "c.\<^sub>FG<f'> = E'\<up>Active(E\<up>ActRef B)"
    from this(3) have "c.\<^sub>FG<f'> = (E'\<circ>(cActive E))\<up>ActRef B" by simp
    with fut `G \<in> dom c` `f' \<in> dom (c.\<^sub>RG)`
    show ?thesis
      by (intro rev_subsetD[OF _ actdom_and_red_inc[OF `c \<rightarrow>\<^sub>\<zeta> c'`]], blast)
  next
    fix G :: ActivityRef and f' :: FutureRef and 
        E' :: general_context and E'' :: general_context and 
        C :: ActivityRef and l :: Label
    assume "G \<in> dom c" "f' \<in> dom (c.\<^sub>RG)" "c.\<^sub>FG<f'> = E'\<up>Upd (ActRef C) l (E''\<up>ActRef B)"
    from this(3) have "c.\<^sub>FG<f'> = (E'\<circ>(cUpdR (ActRef C) l E''))\<up>ActRef B" by simp
    with fut `G \<in> dom c` `f' \<in> dom (c.\<^sub>RG)` show ?thesis
      by (intro rev_subsetD[OF _ actdom_and_red_inc[OF `c \<rightarrow>\<^sub>\<zeta> c'`]], blast)
  qed
next
  fix c :: configuration and c' :: configuration and E :: general_context and
      A :: ActivityRef and B :: ActivityRef and fk :: FutureRef
  assume "c \<rightarrow>\<^sub>\<zeta> c'" and
    act: "\<forall>Aa\<in>dom c. \<forall>(Ea::general_context) Ba. (c.\<^sub>AAa) = Ea\<up>ActRef Ba \<longrightarrow> Ba \<in> dom c" and
    fut: "\<forall>Aa\<in>dom c. \<forall>fk\<in>dom (c.\<^sub>RAa). \<forall>(Ea::general_context) Ba. c.\<^sub>FAa<fk> = Ea\<up>ActRef Ba
           \<longrightarrow> Ba \<in> dom c" and "A \<in> dom c'" "fk \<in> dom (c'.\<^sub>RA)" "c'.\<^sub>FA<fk> = E\<up>ActRef B"
  from aref_and_red_queue[OF this(1) this(4-6)] show "B \<in> dom c'"
  proof (elim disjE conjE exE bexE)
    fix G :: ActivityRef and E' :: general_context and fn :: FutureRef
    assume "G \<in> dom c" "fn \<in> dom (c.\<^sub>RG)" "c.\<^sub>FG<fn> = E'\<up>ActRef B" 
    with fut show ?thesis
      by (simp add: rev_subsetD[OF _ actdom_and_red_inc[OF `c \<rightarrow>\<^sub>\<zeta> c'`]]) 
  next
    fix G :: ActivityRef and E' :: general_context
    assume "G \<in> dom c" "c.\<^sub>AG = E'\<up>ActRef B" 
    with act show ?thesis
      by (simp add: rev_subsetD[OF _ actdom_and_red_inc[OF `c \<rightarrow>\<^sub>\<zeta> c'`]]) 
  next
    assume "created_act c c' B" with created_def[of c c' B] show ?thesis
      by blast
  qed
qed

(* ----------------   W F  2 --------------------  *)
subsection{* WF2 *}

lemma red_fut_00:
  fixes c :: configuration and c' :: configuration and B :: ActivityRef and f :: FutureRef
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "B \<in> dom c" "f \<in> dom (c.\<^sub>RB)"
  shows "\<exists>B\<in>dom c'. f \<in> dom (c'.\<^sub>RB)"
proof -
  from assms(2) futdom_and_red_inc[OF assms(1)]
  have "dom (c.\<^sub>RB) \<subseteq> dom (c'.\<^sub>RB)" by simp
  from subsetD[OF this assms(3)]
       subsetD[OF actdom_and_red_inc[OF assms(1)] assms(2)]
  show ?thesis by blast
qed

(* reduction maintains WF2a+b*)
lemma subject_reduction_WF2[rule_format]: 
  "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> (wf_2a c \<and> wf_2b c \<longrightarrow> wf_2a c' \<and> wf_2b c')"
  unfolding wf_2a_def wf_2b_def
proof (rule impI, elim conjE, intro conjI strip)
  fix c :: configuration and c' :: configuration and E :: general_context and
      A :: ActivityRef and f :: FutureRef
  assume "c \<rightarrow>\<^sub>\<zeta> c'" and
    act: "\<forall>Aa\<in>dom c. \<forall>(Ea::general_context) fa. (c.\<^sub>AAa) = Ea\<up>FutRef fa 
           \<longrightarrow> (\<exists>B\<in>dom c. fa \<in> dom (c.\<^sub>RB))" and
    fut: "\<forall>Aa\<in>dom c. \<forall>fk\<in>dom (c.\<^sub>RAa). \<forall>(Ea::general_context) fa. c.\<^sub>FAa<fk> = Ea\<up>FutRef fa
           \<longrightarrow> (\<exists>B\<in>dom c. fa \<in> dom (c.\<^sub>RB))" and "A \<in> dom c'" "c'.\<^sub>AA = E\<up>FutRef f"
  from ref_and_red_ao[OF this(1) this(4-5)] show "\<exists>B\<in>dom c'. f \<in> dom (c'.\<^sub>RB)"
  proof (simp, elim disjE conjE exE bexE)
    fix A' :: ActivityRef and E' :: general_context
    assume "A' \<in> dom c" "c.\<^sub>AA' = E'\<up>FutRef f" 
    with act obtain B where "B \<in> dom c" "f \<in> dom (c.\<^sub>RB)" by blast
    from red_fut_00[OF `c \<rightarrow>\<^sub>\<zeta> c'` this] show ?thesis by assumption
  next
    fix G :: ActivityRef and f' :: FutureRef and E' :: general_context
    assume "G \<in> dom c" "f' \<in> dom (c.\<^sub>RG)" "c.\<^sub>FG<f'> = E'\<up>Active(E\<up>FutRef f)"
    from this(3) have "c.\<^sub>FG<f'> = (E'\<circ>(cActive E))\<up>FutRef f" by simp
    with fut `G \<in> dom c` `f' \<in> dom (c.\<^sub>RG)` 
    obtain B where "B \<in> dom c" "f \<in> dom(c.\<^sub>RB)" by blast
    from red_fut_00[OF `c \<rightarrow>\<^sub>\<zeta> c'` this] show ?thesis by assumption
  next
    fix G :: ActivityRef and f' :: FutureRef and 
        E' :: general_context and E'' :: general_context and 
        C :: ActivityRef and l :: Label
    assume "G \<in> dom c" "f' \<in> dom (c.\<^sub>RG)" "c.\<^sub>FG<f'> = E'\<up>Upd (ActRef C) l (E''\<up>FutRef f)"
    from this(3) have "c.\<^sub>FG<f'> = (E'\<circ>(cUpdR (ActRef C) l E''))\<up>FutRef f" by simp
    with fut `G \<in> dom c` `f' \<in> dom (c.\<^sub>RG)` 
    obtain B where "B \<in> dom c" "f \<in> dom(c.\<^sub>RB)" by blast
    from red_fut_00[OF `c \<rightarrow>\<^sub>\<zeta> c'` this] show ?thesis by assumption
  qed
next
  fix c :: configuration and c' :: configuration and E :: general_context and
      A :: ActivityRef and f :: FutureRef and fk :: FutureRef
  assume "c \<rightarrow>\<^sub>\<zeta> c'" and
    act: "\<forall>Aa\<in>dom c. \<forall>(Ea::general_context) fa. (c.\<^sub>AAa) = Ea\<up>FutRef fa 
           \<longrightarrow> (\<exists>B\<in>dom c. fa \<in> dom (c.\<^sub>RB))" and
    fut: "\<forall>Aa\<in>dom c. \<forall>fk\<in>dom (c.\<^sub>RAa). \<forall>(Ea::general_context) fa. c.\<^sub>FAa<fk> = Ea\<up>FutRef fa
           \<longrightarrow> (\<exists>B\<in>dom c. fa \<in> dom (c.\<^sub>RB))" and
         "A \<in> dom c'" "fk \<in> dom (c'.\<^sub>RA)" "c'.\<^sub>FA<fk> = E\<up>FutRef f"
  from fref_and_red_queue[OF this(1) this(4-6)] show "\<exists>B\<in>dom c'. f \<in> dom (c'.\<^sub>RB)"
  proof (elim disjE conjE exE bexE)
    fix G :: ActivityRef and E' :: general_context and fn :: FutureRef
    assume "G \<in> dom c" "fn \<in> dom (c.\<^sub>RG)" "c.\<^sub>FG<fn> = E'\<up>FutRef f" 
    with fut obtain B where "B \<in> dom c" "f \<in> dom (c.\<^sub>RB)" by blast
    from red_fut_00[OF `c \<rightarrow>\<^sub>\<zeta> c'` this] show ?thesis by assumption
  next
    fix G :: ActivityRef and E' :: general_context
    assume "G \<in> dom c" "c.\<^sub>AG = E'\<up>FutRef f" with act
    obtain B where "B \<in> dom c" "f \<in> dom (c.\<^sub>RB)" by blast
    from red_fut_00[OF `c \<rightarrow>\<^sub>\<zeta> c'` this] show ?thesis by assumption
  next
    fix B :: ActivityRef
    assume "created_fut c c' B f" with fcreated_def[of c c' B f]
    show ?thesis by auto
  qed
qed

(* reduction maintains well-formedness *)
theorem wf_preserved:
  fixes c :: configuration and c' :: configuration
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "wf_conf c"
  shows "wf_conf c'"
proof -
  from subject_reduction_WF1[OF assms(1)] assms(2)
  have "wf_1a c' \<and> wf_1b c'" by (simp add: wf_conf_def)
  moreover
  from subject_reduction_WF2[OF assms(1)] assms(2)
  have "wf_2a c' \<and> wf_2b c'" by (simp add: wf_conf_def)
  ultimately show ?thesis by (simp add: wf_conf_def)
qed

lemma fold_rec_forw:
  fixes f :: "'a \<Rightarrow> bool \<Rightarrow> bool" and A :: "'a \<Rightarrow> bool" and a :: 'a and z :: bool
  assumes "fun_left_comm f" "finite A" "a \<in> A" "fold f z A"
  shows "f a (fold f z (A - {a}))"
proof -
  from assms(4) fun_left_comm.fold_rec[OF assms(1-3)] show ?thesis by simp
qed

lemma and_fun_left_comm: "fun_left_comm (\<lambda>x y. y \<and> (g x))"
  unfolding fun_left_comm_def by auto

(* now the particular case of a configuration obtain from initconf, 
  and transitive closure-related theorems*)

lemma noRef_act: "\<not>noRef((E::general_context)\<up>ActRef A)"
  by (induct E, auto)

lemma noRef_fut: "\<not>noRef((E::general_context)\<up>FutRef fk)"
  by (induct E, auto)

lemma wf_init_conf: "noRef a \<Longrightarrow> wf_conf (init_config a)"
  unfolding wf_conf_def wf_1a_def wf_1b_def wf_2a_def wf_2b_def init_config_def
            RequestQueue_def OneRequest_def ActiveObject_def
proof (auto simp: noRef_fut noRef_act)
  fix E :: general_context and B :: ActivityRef
  assume "Obj empty (Object empty) = E\<up>ActRef B"
  thus "B = aref 0" proof (case_tac E, auto)
    fix f :: FmapLabel and E' :: general_context
    assume "empty = FLmap f(FLlabel f \<mapsto> E'\<up>ActRef B)"
    thus ?thesis by (simp add: not_sym[OF map_upd_nonempty])
  qed
next
  fix E :: general_context and fk :: FutureRef
  assume "Obj empty (Object empty) = E\<up>FutRef fk"
  thus "fk = fref 0" proof (case_tac E, auto)
    fix f :: FmapLabel and E' :: general_context
    assume "empty = FLmap f(FLlabel f \<mapsto> E'\<up>FutRef fk)"
    thus ?thesis by (simp add: not_sym[OF map_upd_nonempty])
  qed
qed

lemma wf_conf_rtrancl[rule_format]: "c \<rightarrow>\<^sub>\<zeta>\<^sup>* c' \<Longrightarrow> wf_conf c \<longrightarrow> wf_conf c'"
proof (induct rule: rtranclp.induct, simp, intro impI, simp)
  fix b :: configuration and c :: configuration
  assume "b \<rightarrow>\<^sub>\<zeta> c" "wf_conf b" from wf_preserved[OF this]
  show "wf_conf c" by assumption
qed

lemma wf_reachable: "c \<in> reachable_configs \<Longrightarrow> wf_conf c"
  unfolding reachable_configs_def
proof (simp, elim exE conjE)
  fix a :: sterm and c :: configuration
  assume "noRef a" "init_config a \<rightarrow>\<^sub>\<zeta>\<^sup>* c"
  from wf_conf_rtrancl[OF this(2) wf_init_conf[OF this(1)]]
  show "wf_conf c" by assumption
qed

section {* Disjunct request queue domains *}
(* disjunuct futures is true if no two request correspond to the same future 
  (should be the case for a reasonable conf) *)
constdefs 
  disjunct_futures:: "configuration \<Rightarrow> bool"
  "disjunct_futures c == (\<forall>A\<in>dom c. \<forall>B\<in>dom c. A \<noteq> B \<longrightarrow> (dom (c.\<^sub>RA) \<inter> dom (c.\<^sub>RB)) = {})"

(* reduction maintains disjunctness of futures *)
lemma disjunct_futures_SR:
  fixes c :: configuration and c' :: configuration
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "disjunct_futures c"
  shows "disjunct_futures c'"
proof -
  {
    fix ca :: configuration and cb :: configuration and cc :: configuration and
        A :: ActivityRef and fa :: FutureRef and t :: sterm and ma :: futmap and a :: sterm
    assume "disjunct_futures cc" "cc = ca" and
   def_cb: "cb = ca(A \<mapsto> (ma(fa \<mapsto> t),a))" and
           "A \<in> dom ca" "fa \<in> dom (ca.\<^sub>RA)" "dom ma = dom (ca.\<^sub>RA)"
    hence "disjunct_futures cb"
      unfolding disjunct_futures_def proof (intro strip)
      fix A' :: ActivityRef and B' :: ActivityRef
      assume "A' \<in> dom cb" "B' \<in> dom cb" "A' \<noteq> B'"
      with `disjunct_futures cc` `cc = ca` def_cb
	   `A \<in> dom ca` `A' \<in> dom cb` `B' \<in> dom cb` `A' \<noteq> B'` 
      have "dom (ca.\<^sub>RA') \<inter> dom (ca.\<^sub>RB') = {}"
	unfolding disjunct_futures_def by auto
      moreover
      from def_cb `A \<in> dom ca` 
	   `fa \<in> dom (ca.\<^sub>RA)` `dom ma = dom (ca.\<^sub>RA)` 
      have "\<forall>B\<in>dom cb. dom (cb.\<^sub>RB) = dom (ca.\<^sub>RB)"
	by (auto simp: RequestQueue_def)
      with `A' \<in> dom cb` `B' \<in> dom cb` 
      have "dom (cb.\<^sub>RA') = dom (ca.\<^sub>RA') \<and> dom (cb.\<^sub>RB') = dom (ca.\<^sub>RB')"
	by auto
      ultimately show "dom (cb.\<^sub>RA') \<inter> dom (cb.\<^sub>RB') = {}" by simp
    qed
  }
  note lem1 = this

  {
    fix ca :: configuration and cb :: configuration and cc :: configuration and
        A :: ActivityRef and fa :: FutureRef and t :: sterm and ma :: futmap and a :: sterm
    assume "disjunct_futures cc" "cc = ca" and
   def_cb: "cb = ca(A \<mapsto> (ma(fa \<mapsto> t),a))" and
           "A \<in> dom ca" "\<forall>A\<in>dom ca. fa \<notin> dom (ca.\<^sub>RA)" "dom ma = dom (ca.\<^sub>RA)"
    hence "disjunct_futures cb"
      unfolding disjunct_futures_def proof (intro strip)
      fix A' :: ActivityRef and B' :: ActivityRef
      assume "A' \<in> dom cb" "B' \<in> dom cb" "A' \<noteq> B'"
      with prems(3-6) have "dom (ca.\<^sub>RA') \<inter> dom (ca.\<^sub>RB') = {}"
	unfolding disjunct_futures_def by auto
      show "dom (cb.\<^sub>RA') \<inter> dom (cb.\<^sub>RB') = {}" proof (cases "A' = A")
	case True with `cb = ca(A \<mapsto> (ma(fa \<mapsto> t), a))` `A \<in> dom ca` 
	               `dom ma = dom (ca.\<^sub>RA)` 
	have "dom (cb.\<^sub>RA') = dom(ca.\<^sub>RA') \<union> {fa}"
	  by (auto simp: RequestQueue_def)
	moreover
	from def_cb `\<forall>A\<in>dom ca. fa \<notin> dom (ca.\<^sub>RA)` `B' \<in> dom cb` `A' \<noteq> B'` `A' = A` 
	have "fa \<notin> dom (ca.\<^sub>RB')" by simp
	moreover
	from def_cb `dom ma = dom (ca.\<^sub>RA)` `A' \<noteq> B'` `A' = A` 
	have "dom (cb.\<^sub>RB') = dom(ca.\<^sub>RB')"
	  by (auto simp: RequestQueue_def)
	ultimately show ?thesis using `dom (ca.\<^sub>RA') \<inter> dom (ca.\<^sub>RB') = {}` by simp
      next
	case False with def_cb `\<forall>A\<in>dom ca. fa \<notin> dom (ca.\<^sub>RA)` `A' \<in> dom cb` 
	have "fa \<notin> dom (ca.\<^sub>RA')" by simp
	moreover
	from def_cb `dom ma = dom (ca.\<^sub>RA)` `A' \<noteq> A`
	have "dom (cb.\<^sub>RA') = dom (ca.\<^sub>RA')"
	  by (auto simp: RequestQueue_def)
	moreover
	from def_cb `dom ma = dom (ca.\<^sub>RA)` `B' \<in> dom cb`
	have "dom (cb.\<^sub>RB') \<subseteq> dom (ca.\<^sub>RB') \<union> {fa}"
	  by (auto simp: RequestQueue_def)
	ultimately show ?thesis using `dom (ca.\<^sub>RA') \<inter> dom (ca.\<^sub>RB') = {}` by blast
      qed
    qed
  }
  note lem2 = this

  {
    fix ca :: configuration and cb :: configuration and cc :: configuration and
        A :: ActivityRef and fa :: FutureRef and t :: sterm and ma :: futmap and 
        a :: sterm and B :: ActivityRef and b :: sterm
    assume "disjunct_futures cc" "cc = ca" and
   def_cb: "cb = ca(A \<mapsto> (ma(fa \<mapsto> t),a),B \<mapsto> (empty,b))"  and
           "B \<notin> dom ca" "A \<in> dom ca" "fa \<in> dom (ca.\<^sub>RA)" "dom ma = dom (ca.\<^sub>RA)"
    hence "disjunct_futures cb"
      unfolding disjunct_futures_def proof (intro strip)
      def ce \<equiv> "ca(A \<mapsto> (ma(fa \<mapsto> t),a))"
      note lem1[OF `disjunct_futures cc` `cc = ca` _ `A \<in> dom ca` 
	           `fa \<in> dom (ca.\<^sub>RA)` `dom ma = dom (ca.\<^sub>RA)`]
      from this[of ce t a] have "disjunct_futures ce" by (simp add: ce_def)
      moreover
      {
	fix D :: ActivityRef assume "D \<in> dom cb"
	with def_cb `dom ma = dom (ca.\<^sub>RA)` 
	have "dom (cb.\<^sub>RD) = {} \<or> (D \<in> dom ce \<and> dom (cb.\<^sub>RD) = dom (ce.\<^sub>RD))"
	 by (auto simp: ce_def RequestQueue_def)
      }
      note domlem = this
      fix A' :: ActivityRef and B' :: ActivityRef
      assume "A' \<in> dom cb" "B' \<in> dom cb" "A' \<noteq> B'"
      note this(3) domlem[OF this(1)] domlem[OF this(2)]
      ultimately show "dom (cb.\<^sub>RA') \<inter> dom (cb.\<^sub>RB') = {}"
	unfolding disjunct_futures_def
	by (elim disjE, auto)
    qed
  }
  note lem3 = this

(* case split *)
  from assms show ?thesis
  proof (cases rule: zeta.cases)
    case (zeta_local A ca fk a')
    from lem1[OF `disjunct_futures c` `c = ca` 
                 `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> a'),ca.\<^sub>AA))` 
                 `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)`] 
    show ?thesis by simp
  next
    case (reply A ca fa Ea fn B)
    from lem1[OF `disjunct_futures c` `c = ca` 
                 `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> Ea\<up>(ca.\<^sub>FB<fn>)),ca.\<^sub>AA))` 
                 `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)`] 
    show ?thesis by simp
  next
    case (active G ca A fk E a) 
    from lem3[OF `disjunct_futures c` `c = ca` 
                 `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>ActRef G),ca.\<^sub>AA),
                          G \<mapsto> (empty,a))` 
                 `G \<notin> dom ca` `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)`]
    show ?thesis by simp
  next
    case (update_AO G ca A B fa E l p) 
    from lem3[OF `disjunct_futures c` `c = ca` 
                 `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> E\<up>ActRef G),ca.\<^sub>AA),
                          G \<mapsto> (empty,Upd (ca.\<^sub>AB) l p))` 
                 `G \<notin> dom ca` `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)`]
    show ?thesis by simp
  next
    case (request ca fb A B fa E l p)
    def cb \<equiv> "ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> E\<up>FutRef fb),ca.\<^sub>AA))"
    from `B \<in> dom ca` have "B \<in> dom cb" by (simp add: cb_def)
    from `\<forall>A\<in> dom ca. fb \<notin> dom (ca.\<^sub>RA)` `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)`
    have fb: "\<forall>A\<in>dom cb. fb \<notin> dom (cb.\<^sub>RA)"
      by (auto simp: cb_def RequestQueue_def)
    note lem1[OF `disjunct_futures c` `c = ca` _ `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)`]
    from this[of cb "ca.\<^sub>RA" "E\<up>FutRef fb" "ca.\<^sub>AA"]
    have "disjunct_futures cb" by (simp add: cb_def)
    note lem2[OF this _ _ `B \<in> dom cb` fb]
    from this[of c' "ca.\<^sub>RB" "Call (ca.\<^sub>AB) l p" "ca.\<^sub>AB"] 
         `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> E\<up>FutRef fb),ca.\<^sub>AA),
                  B \<mapsto> ((ca.\<^sub>RB)(fb \<mapsto> Call (ca.\<^sub>AB) l p),ca.\<^sub>AB))`
         `A \<noteq> B`
    show ?thesis by (auto simp: cb_def RequestQueue_def)
  next
    case (request_self ca fn A fa E l p)
    def cb \<equiv> "ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> E\<up>FutRef fn),ca.\<^sub>AA))"
    from `A \<in> dom ca` have "A \<in> dom cb" by (simp add: cb_def)
    from `\<forall>A\<in>dom ca. fn \<notin> dom (ca.\<^sub>RA)` `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)` 
    have fn: "\<forall>A\<in>dom cb. fn \<notin> dom (cb.\<^sub>RA)"
      by (auto simp: cb_def RequestQueue_def)
    note lem1[OF `disjunct_futures c` `c = ca` _ `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)`]
    from this[of cb "ca.\<^sub>RA" "E\<up>FutRef fn" "ca.\<^sub>AA"]
    have "disjunct_futures cb" by (simp add: cb_def)
    note lem2[OF this _ _ `A \<in> dom cb` fn]
    from this[of c' "(ca.\<^sub>RA)(fa \<mapsto> E\<up>FutRef fn)" "Call (ca.\<^sub>AA) l p" "ca.\<^sub>AA"]
         `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> E\<up>FutRef fn, fn \<mapsto> Call (ca.\<^sub>AA) l p),ca.\<^sub>AA))`
         `fa \<in> dom (ca.\<^sub>RA)`
    show ?thesis by (auto simp: cb_def RequestQueue_def)
  qed
qed

lemma initial_config_disjunct: "disjunct_futures (init_config a)"
  unfolding disjunct_futures_def init_config_def by auto

(* transitive closure extension *)
lemma disjunct_rtrancl[rule_format]: 
  "c \<rightarrow>\<^sub>\<zeta>\<^sup>* c' \<Longrightarrow> disjunct_futures c \<longrightarrow> disjunct_futures c'"
proof (induct rule: rtranclp.induct, simp, intro impI, simp)
  fix a :: configuration and b :: configuration and c :: configuration
  assume "b \<rightarrow>\<^sub>\<zeta> c" "disjunct_futures b"
  from  disjunct_futures_SR[OF this] show "disjunct_futures c" by assumption
qed

(* config reachable from initconf => disjunct futures *)
theorem reachable_config_disjunct: "c \<in> reachable_configs \<Longrightarrow> disjunct_futures c"
  unfolding reachable_configs_def
proof (simp+, elim exE conjE)
  fix a :: sterm assume "init_config a \<rightarrow>\<^sub>\<zeta>\<^sup>* c"
  from disjunct_rtrancl[OF this initial_config_disjunct[of a]]
  show ?thesis by assumption
qed

end